""" Unit tests for spikes.py module """
import pytest
import numpy as np

from spynal.tests.data_fixtures import MISSING_ARG_ERRS
from spynal.utils import iarange, unsorted_unique, setup_sliding_windows, \
                         object_array_equal, concatenate_object_array
from spynal.spikes import simulate_spike_trains, simulate_spike_waveforms, \
                          times_to_bool, bool_to_times, \
                          cut_trials, select_time_range, realign_data, pool_electrode_units, \
                          rate, rate_stats, isi, isi_stats, waveform_stats, \
                          plot_raster, plot_mean_waveforms, plot_waveform_heatmap
from spynal.spectra.multitaper import compute_tapers


# =============================================================================
# Fixtures for generating simulated data
# =============================================================================
@pytest.fixture(scope='session')
def spike_timestamp():
    """
    Fixture simulates set of spike trains as spike timestamps to use for all unit tests

    RETURNS
    data    (10,2) ndarray of (n_spikes,) objects. Simulated spike timestamps
            (eg simulating 10 trials x 2 units)

    timepts None. Second argout returned only to match output of spike_bool
    """
    # Note: seed=1 makes data reproducibly match output of Matlab
    data, _ = simulate_spike_trains(n_trials=20, n_conds=1, refractory=1e-3, seed=1)
    # Reshape data to simulate (n_trials,n_units) spiking data
    return data.reshape(10,2), None


@pytest.fixture(scope='session')
def spike_bool(spike_timestamp):
    """
    Fixture simulates set of spike trains as binary trains to use for all unit tests

    RETURNS
    data    (10,2,1001) ndarray of bool. Simulated binary spike trains
            (eg simulating 10 trials x 2 units x 1001 timepts).

    timepts (1001,) ndarray of float. Time sampling vector for data (in s).
    """
    # Note: Implicitly also tests times_to_bool function
    return times_to_bool(spike_timestamp[0], lims=[0,1])


@pytest.fixture(scope='session')
def spike_data(spike_timestamp, spike_bool):
    """
    "Meta-fixture" that returns both timestamp and boolean spiking data types
    in a dictionary.

    RETURNS
    data_dict   {'data_type' : (data,timepts)} dict containing outputs from
                each of constituent fixtures

    SOURCE      https://stackoverflow.com/a/42400786
    """
    return {'spike_timestamp': spike_timestamp, 'spike_bool': spike_bool}


@pytest.fixture(scope='session')
def spike_timestamp_trial_uncut(spike_timestamp):
    """
    Simulates set of spike trains as spike timestamps not cut into trials

    RETURNS
    data    (2,) ndarray of (n_spikes,) objects. Simulated spike timestamps
            (eg simulating 2 units, uncut into trials)

    timepts None. Second argout returned only to match output of spike_bool
    """
    data, _ = spike_timestamp
    n_trials,n_units = data.shape
    data_uncut = np.empty((n_units,),dtype=object)
    for unit in range(n_units):
        data_uncut[unit] = np.hstack([data[trial,unit]+trial for trial in range(n_trials)])

    return data_uncut, None


@pytest.fixture(scope='session')
def spike_bool_trial_uncut(spike_bool):
    """
    Fixture simulates set of spike trains as binary trains to use for all unit tests

    RETURNS
    data    (2,1001) ndarray of bool. Simulated binary spike trains
            (eg simulating 2 units x 1001 timepts, uncut into trials).

    timepts (1001,) ndarray of float. Time sampling vector for data (in s).
    """
    data, timepts = spike_bool
    n_trials,n_units,n_timepts = data.shape
    data = data.transpose((1,2,0)).reshape((n_units,n_timepts*n_trials),order='F')
    timepts = np.hstack([timepts + trial*n_timepts for trial in range(n_trials)])
    return data, timepts


@pytest.fixture(scope='session')
def spike_data_trial_uncut(spike_timestamp_trial_uncut, spike_bool_trial_uncut):
    """
    "Meta-fixture" that returns both timestamp and boolean spiking data types
    in a dictionary.

    RETURNS
    data_dict   {'data_type' : (data,timepts)} dict containing outputs from
                each of constituent fixtures

    SOURCE      https://stackoverflow.com/a/42400786
    """
    return {'spike_timestamp': spike_timestamp_trial_uncut,
            'spike_bool': spike_bool_trial_uncut}


@pytest.fixture(scope='session')
def spike_waveform(spike_timestamp):
    """
    Fixture simulates spike waveforms of set of spike trains to use for all unit tests

    RETURNS
    data    (10,2) ndarray of (n_timepts,n_spikes) objects. Simulated spike waveforms
            (eg simulating 10 trials x 2 units)

    timepts (49,) ndarray of float. Time sampling vector for waveform data (in s).
    """
    spike_timestamp = spike_timestamp[0]
    n_trials,n_units = spike_timestamp.shape

    spike_waves = np.empty((n_trials,n_units), dtype=object)
    for i_trial in range(n_trials):
        for i_unit in range(n_units):
            n_spikes = len(spike_timestamp[i_trial,i_unit])
            if n_spikes != 0:
                spike_waves[i_trial,i_unit],timepts = simulate_spike_waveforms(n_spikes=n_spikes)
            else:
                spike_waves[i_trial,i_unit] = np.asarray([])

    return spike_waves, timepts


# =============================================================================
# Unit tests for rate computation functions
# =============================================================================
@pytest.mark.parametrize('data_type, output, result',
                         [('spike_timestamp', 'count', 105),
                          ('spike_timestamp', 'rate', 5.25),
                          ('spike_bool', 'count', 105),
                          ('spike_bool', 'rate', 5.25)])
def test_bin_rate(spike_data, data_type, output, result):
    """ Unit tests for bin_rate function for computing binned spike rates """
    # Extract given data type from data dict
    data, timepts = spike_data[data_type]
    data_orig = data.copy()

    if output == 'rate':    dtype_type = float
    elif output == 'count': dtype_type = np.integer
    else:                   dtype_type = bool

    data_checker = np.array_equal if data_type == 'bool' else object_array_equal
    result_checker = np.mean if output == 'rate' else np.sum

    # Basic test of shape, dtype, value of output
    # Test values summed over entire array -> scalar for spike counts
    # Test values averaged over entire array -> scalar for spike rates
    rates, bins = rate(data, method='bin', lims=[0,1], output=output, axis=-1, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert bins.shape == (20, 2)
    assert rates.shape == (10, 2, 20)
    assert np.issubdtype(rates.dtype,dtype_type)
    assert np.isclose(result_checker(rates), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with different data array shape
    shape = (5,2,2,*data.shape[2:])
    rates, bins = rate(data.reshape(shape), method='bin', lims=[0,1], output=output,
                       axis=-1, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert bins.shape == (20, 2)
    assert rates.shape == (5, 2, 2, 20)
    assert np.isclose(result_checker(rates), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed data dimensionality
    # Note: output dims are expected to be different for timestamp vs boolean data
    expected_shape = (20, 2, 10) if data_type == 'spike_bool' else (2, 10, 20)
    rates, bins = rate(data.transpose(), method='bin', lims=[0,1], output=output,
                       axis=0, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert bins.shape == (20, 2)
    assert rates.shape == expected_shape
    assert np.isclose(result_checker(rates), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with different sliding window length
    rates, bins = rate(data, method='bin', lims=[0,1], width=20e-3, output=output,
                       axis=-1, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert bins.shape == (50, 2)
    assert rates.shape == (10, 2, 50)
    assert np.isclose(result_checker(rates), result, rtol=1e-2, atol=1e-2)

    # Test for consistent ouptut when bins are set explicitly
    bins = setup_sliding_windows(20e-3,[0,1])
    rates, bins = rate(data, method='bin', bins=bins, output=output, axis=-1, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert bins.shape == (50, 2)
    assert rates.shape == (10, 2, 50)
    assert np.isclose(result_checker(rates), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output for custom unequal-width bins
    # Note: Summed counts should be same, but mean rates are expected to be slightly different here
    bins = [[0,250e-3], [250e-3,750e-3], [750e-3,1000e-3]]
    rates, bins = rate(data, method='bin', bins=bins, output=output, axis=-1, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert rates.shape == (10, 2, 3)
    if output == 'rate':
        assert np.isclose(rates.mean(), 5.53, rtol=1e-2, atol=1e-2)
    else:
        assert np.isclose(result_checker(rates), result, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        rates, bins = rate(data, method='bin', lims=[0,1], output=output, axis=-1,
                           timepts=timepts, foo=None)


@pytest.mark.parametrize('data_type, kernel, result',
                         [('spike_timestamp', 'gaussian', 4.92),
                          ('spike_timestamp', 'hanning', 4.93),
                          ('spike_bool', 'gaussian', 5.25),
                          ('spike_bool', 'hanning', 5.25)])
def test_density(spike_data, data_type, kernel, result):
    """ Unit tests for density function for computing spike densities """
    # Extract given data type from data dict
    data, timepts = spike_data[data_type]
    data_orig = data.copy()

    # Set kernel width parameter so gaussian and hanning kernels are ~ identical
    width = 50e-3 if kernel == 'gaussian' else 125e-3

    data_checker = np.array_equal if data_type == 'bool' else object_array_equal

    # Basic test of shape, dtype, value of output
    # Test values averaged over entire array -> scalar for spike rates
    rates, tout = rate(data, method='density', kernel=kernel, width=width, lims=[0,1],
                       axis=-1, timepts=timepts)
    print(rates.mean(), tout.shape, rates.shape)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert tout.shape == (1001,)
    assert rates.shape == (10, 2, 1001)
    assert np.isclose(rates.mean(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with different data array shape
    shape = (5,2,2,*data.shape[2:])
    rates, tout = rate(data.reshape(shape), method='density', kernel=kernel, width=width,
                       lims=[0,1], axis=-1, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert tout.shape == (1001,)
    assert rates.shape == (5, 2, 2, 1001)
    assert np.isclose(rates.mean(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed data dimensionality
    # Note: output dims are expected to be different for timestamp vs boolean data
    expected_shape = (1001, 2, 10) if data_type == 'spike_bool' else (2, 10, 1001)
    rates, tout = rate(data.transpose(), method='density', kernel=kernel, width=width,
                       lims=[0,1], axis=0, timepts=timepts)
    print(rates.mean(), tout.shape, rates.shape)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert tout.shape == (1001,)
    assert rates.shape == expected_shape
    assert np.isclose(rates.mean(), result, rtol=1e-2, atol=1e-2)

    # Test for ~ consistent ouptut with 10x downsampling after spike density estimation
    rates, tout = rate(data, method='density', kernel=kernel, width=width, lims=[0,1],
                       step=10e-3, axis=-1, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert tout.shape == (101,)
    assert rates.shape == (10, 2, 101)
    assert np.isclose(rates.mean(), result, rtol=0.01, atol=0.01)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        rates, tout = rate(data, method='density', kernel=kernel, width=width, lims=[0,1],
                           axis=-1, timepts=timepts, foo=None)


# =============================================================================
# Unit tests for spike rate, ISI, and waveform stats functions
# =============================================================================
@pytest.mark.parametrize('data_type, stat, result',
                         [('spike_timestamp', 'Fano', 42.0),
                          ('spike_bool', 'CV', 2.29)])
def test_rate_stats(spike_data, data_type, stat, result):
    """ Unit tests for rate_stats function for computing spike rate statistics """
    # Extract given data type from data dict
    data, timepts = spike_data[data_type]

    # Compute spike rates from timestamp/binary spike data -> (n_trials,n_unit,n_timepts)
    rates, _ = rate(data, method='bin', lims=[0,1], output='rate', axis=-1, timepts=timepts)
    rates_orig = rates.copy()
    n_trials,n_chnls,n_timepts = rates.shape

    data_checker = np.array_equal if data_type == 'bool' else object_array_equal

    # Basic test of shape, value of output
    # Test value of 1st trial/channel as exemplar
    stats = rate_stats(rates, stat=stat, axis=0)
    assert data_checker(rates,rates_orig)     # Ensure input data not altered by func
    assert stats.shape == (1, n_chnls, n_timepts)
    assert np.isclose(stats[0,0,0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with different data array shape (additional axis inserted)
    stats = rate_stats(np.concatenate((rates[:,:,np.newaxis,:],rates[:,:,np.newaxis,:]), axis=2),
                       stat=stat, axis=0)
    assert data_checker(rates,rates_orig)     # Ensure input data not altered by func
    assert stats.shape == (1, n_chnls, 2, n_timepts)
    assert np.isclose(stats[0,0,0,0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed data dimensionality
    stats = rate_stats(rates.T, stat=stat, axis=-1)
    assert data_checker(rates,rates_orig)     # Ensure input data not altered by func
    assert stats.shape == (n_timepts, n_chnls, 1)
    assert np.isclose(stats[0,0,0], result, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        stats = rate_stats(rates, stat=stat, axis=0, foo=None)


@pytest.mark.parametrize('data_type, stat, result',
                         [('spike_timestamp', 'Fano', 0.1006),
                          ('spike_bool', 'CV', 0.9110),
                          ('spike_bool', 'CV2', 0.9867),
                          ('spike_bool', 'LV', 0.9423),
                          ('spike_bool', 'burst_fract', 0.2041)])
def test_isi_stats(spike_data, data_type, stat, result):
    """ Unit tests for isi_stats function for computing inter-spike interval statistics """
    # Extract given data type from data dict
    data, timepts = spike_data[data_type]
    data_orig = data.copy()

    data_checker = np.array_equal if data_type == 'bool' else object_array_equal

    # Compute inter-spike intervals from timestamp/binary spike data -> (n_trials,n_units)
    ISIs = isi(data, axis=-1, timepts=timepts)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    n_trials,n_chnls = ISIs.shape

    # KLUDGE Pool ISI data across trials for locality-sensitive metrics, so output shape is same
    if stat in ['CV2','LV']: ISIs = concatenate_object_array(ISIs,axis=0,sort=False)
    ISIs_orig = ISIs.copy()

    axis = 'each' if stat in ['CV2','LV'] else 0

    # Basic test of shape, value of output
    # Test value of 1st trial/channel as exemplar
    stats = isi_stats(ISIs, stat=stat, axis=axis)
    assert data_checker(ISIs,ISIs_orig)     # Ensure input data not altered by func
    assert stats.shape == (1, n_chnls)
    assert np.isclose(stats[0,0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with different data array shape (additional axis inserted)
    stats = isi_stats(np.concatenate((ISIs[:,:,np.newaxis],ISIs[:,:,np.newaxis]), axis=2),
                      stat=stat, axis=axis)
    assert data_checker(ISIs,ISIs_orig)     # Ensure input data not altered by func
    assert stats.shape == (1, n_chnls, 2)
    assert np.isclose(stats[0,0,0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed data dimensionality
    stats = isi_stats(ISIs.T, stat=stat, axis='each' if stat in ['CV2','LV'] else -1)
    assert data_checker(ISIs,ISIs_orig)     # Ensure input data not altered by func
    assert stats.shape == (n_chnls, 1)
    assert np.isclose(stats[0,0], result, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        ISIs = isi(data, axis=-1, timepts=timepts, foo=None)
        stats = isi_stats(ISIs, stat=stat, axis=axis, foo=None)


@pytest.mark.parametrize('stat, result',
                         [('width',         433.33),
                          ('repolarization',200.00),
                          ('trough_width',  166.67),
                          ('amp_ratio',     1.55)])
def test_waveform_stats(spike_waveform, stat, result):
    """ Unit tests for waveform_stats function for computing spike waveform statistics """
    # Extract given data type from data dict
    data,timepts = spike_waveform
    n_trials,n_units = data.shape
    n_timepts,n_spikes0 = data[0,0].shape
    data_orig = data.copy()

    extra_args = dict(smp_rate=30e3) if stat != 'amp_ratio' else {}
    result_mult = 1 if stat == 'amp_ratio' else 1e6 # Convert temporal stats from s -> us

    # Basic test of shape, value of output
    # Test value of 1st trial/channel as exemplar
    stats = waveform_stats(data, stat=stat, axis=0, **extra_args)
    print(stats[0,0].shape, np.round(stats[0,0][0,0]*result_mult,2))
    assert object_array_equal(data,data_orig)     # Ensure input data not altered by func
    assert stats.shape == (n_trials,n_units)
    assert stats[0,0].shape == (1,n_spikes0)
    assert np.isclose(stats[0,0][0,0]*result_mult, result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with non-object array inputs (just ndarray for single trial,unit)
    trial_data = data[0,0]
    trial_data_orig = trial_data.copy()
    stats = waveform_stats(trial_data, stat=stat, axis=0, **extra_args)
    assert np.array_equal(trial_data,trial_data_orig)     # Ensure input data not altered by func
    assert stats.shape == (1, n_spikes0)
    assert np.isclose(stats[0,0]*result_mult, result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed data dimensionality
    stats = waveform_stats(trial_data.T, stat=stat, axis=-1, **extra_args)
    assert np.array_equal(trial_data,trial_data_orig)     # Ensure input data not altered by func
    assert stats.shape == (n_spikes0, 1)
    assert np.isclose(stats[0,0]*result_mult, result, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        stats = waveform_stats(data, stat=stat, axis=0, foo=None)


# =============================================================================
# Unit tests for rate preprocessing/utility functions
# =============================================================================
def test_bool_to_times(spike_timestamp, spike_bool):
    """ Unit tests for bool_to_times function to convert spike timestamps to binary trains """
    data_timestamp, _   = spike_timestamp
    data_bool, timepts  = spike_bool
    data_orig = data_bool.copy()

    # Test that timestamp->bool->timestamp data retains same dtype, shape, values
    data_bool_to_timestamp = bool_to_times(data_bool, timepts, axis=-1)
    assert np.array_equal(data_bool,data_orig)  # Ensure input data not altered by func
    assert data_bool_to_timestamp.dtype == data_timestamp.dtype
    assert data_bool_to_timestamp.shape == data_timestamp.shape
    assert np.asarray([d1.shape == d2.shape for d1,d2
                       in zip(data_bool_to_timestamp.flatten(), data_timestamp.flatten())]).all()
    assert np.asarray([np.allclose(d1, d2, rtol=1e-3, atol=1e-3) for d1,d2
                       in zip(data_bool_to_timestamp.flatten(), data_timestamp.flatten())]).all()

    # Test for correct handling of single spike trains
    assert np.allclose(bool_to_times(data_bool[0,0,:], timepts, axis=-1), data_timestamp[0,0],
                       rtol=1e-2, atol=1e-2)
    assert np.array_equal(data_bool,data_orig)  # Ensure input data not altered by func


def test_times_to_bool(spike_timestamp, spike_bool):
    """ Unit tests for times_to_bool function to convert binary spike trains to timestamps """
    data_timestamp, _   = spike_timestamp
    data_bool, timepts  = spike_bool
    data_orig = data_timestamp.copy()

    # Test that bool->timestamp->bool data retains same shape, dtype, values
    data_timestamp_to_bool, timepts2 = times_to_bool(data_timestamp, lims=(0,1))
    assert object_array_equal(data_timestamp,data_orig)  # Ensure input data not altered by func
    assert (timepts == timepts2).all()
    assert data_timestamp_to_bool.shape == data_bool.shape
    assert data_timestamp_to_bool.dtype == data_bool.dtype
    assert (data_timestamp_to_bool == data_bool).all()

    # Test for correct handling of single spike trains and list-valued data
    assert (times_to_bool(data_timestamp[0,0], lims=(0,1))[0] == data_bool[0,0,:]).all()
    assert object_array_equal(data_timestamp,data_orig)  # Ensure input data not altered by func
    assert (times_to_bool(list(data_timestamp[0,0]), lims=(0,1))[0] == data_bool[0,0,:]).all()
    assert object_array_equal(data_timestamp,data_orig)  # Ensure input data not altered by func


@pytest.mark.parametrize('data_type',
                         [('spike_timestamp'),('spike_bool')])
def test_cut_trials(spike_data_trial_uncut, spike_data, data_type):
    """ Unit tests for cut_trials function """
    data, _ = spike_data[data_type]
    uncut_data, _ = spike_data_trial_uncut[data_type]
    data_orig = uncut_data.copy()
    n_trials, n_units = data.shape[:2]

    if data_type == 'spike_timestamp':
        trial_lims = np.asarray([0,1])[np.newaxis,:] + np.arange(n_trials)[:,np.newaxis]
        cut_data = cut_trials(uncut_data, trial_lims, trial_refs=np.arange(0,n_trials)).T
        assert object_array_equal(uncut_data,data_orig)     # Ensure input data not altered by func
        for trial in range(n_trials):
            for unit in range(n_units):
                assert np.allclose(cut_data[trial,unit], data[trial,unit])

        # Ensure that passing a nonexistent/misspelled kwarg raises an error
        with pytest.raises(MISSING_ARG_ERRS):
            cut_data = cut_trials(uncut_data, trial_lims, trial_refs=np.arange(0,n_trials), foo=None).T

    else:
        trial_lims = np.asarray([0,1])[np.newaxis,:] + 1.001*np.arange(n_trials)[:,np.newaxis]
        cut_data = cut_trials(uncut_data, trial_lims, smp_rate=1000, axis=1).transpose((2,0,1))
        assert np.array_equal(uncut_data,data_orig)     # Ensure input data not altered by func
        assert (cut_data == data).all()

        # Ensure that passing a nonexistent/misspelled kwarg raises an error
        with pytest.raises(MISSING_ARG_ERRS):
            cut_data = cut_trials(uncut_data, trial_lims, smp_rate=1000, axis=1, foo=None).transpose((2,0,1))

    assert cut_data.shape == data.shape


@pytest.mark.parametrize('data_type, result',
                         [('spike_timestamp',   44),
                          ('spike_bool',        44)])
def test_select_time_range(spike_data, data_type, result):
    """ Unit tests for select_time_range function """
    data, timepts = spike_data[data_type]
    data_orig = data.copy()
    n_trials, n_units = data.shape[:2]
    if data_type == 'spike_bool': n_timepts = data.shape[2]
    time_range = (0.25,0.75)
    n_timepts_out = len(iarange(250,750))

    def check_results(sel_data, tbool):
        if data_type == 'spike_timestamp':
            n_spikes, n_spikes2 = 0, 0
            for trial in range(n_trials):
                for unit in range(n_units):
                    n_spikes += len(sel_data[trial,unit])
                    n_spikes2 += tbool[trial,unit].sum()
                    assert np.all((sel_data[trial,unit] >= time_range[0]) &
                                (sel_data[trial,unit] <= time_range[1]))
            assert n_spikes == n_spikes2

        else:
            n_spikes = sel_data.sum()

        return n_spikes

    # Realign timestamps, then realign back to original timebase and test if same
    if data_type == 'spike_timestamp':
        sel_data, tbool = select_time_range(data, time_range)
        assert object_array_equal(data,data_orig)     # Ensure input data not altered by func
        assert sel_data.shape == data.shape
        assert tbool.shape == data.shape
        assert check_results(sel_data, tbool) == result

        # Test for consistent output with transposed data dimensionality
        sel_data,tbool = select_time_range(data.T, time_range)
        assert sel_data.shape == data.T.shape
        assert tbool.shape == data.T.shape
        assert check_results(sel_data.T, tbool.T) == result

        # Ensure that passing a nonexistent/misspelled kwarg raises an error
        with pytest.raises(MISSING_ARG_ERRS):
            sel_data, tbool = select_time_range(data, time_range, foo=None)

    # For boolean data, realign to 2 distinct times, then concatenate together and test if same
    else:
        sel_data, tbool = select_time_range(data, time_range, time_axis=-1, timepts=timepts)
        assert object_array_equal(data,data_orig)     # Ensure input data not altered by func
        assert sel_data.shape == (*(data.shape[:-1]), n_timepts_out)
        assert tbool.shape == (n_timepts,)
        assert check_results(sel_data, tbool) == result

        # Test for consistent output with transposed data dimensionality
        sel_data, tbool = select_time_range(data.T, time_range, time_axis=0, timepts=timepts)
        assert object_array_equal(data,data_orig)     # Ensure input data not altered by func
        assert sel_data.shape == (n_timepts_out, *(data.T.shape[1:]))
        assert tbool.shape == (n_timepts,)
        assert check_results(sel_data.T, tbool) == result

        # Ensure that passing a nonexistent/misspelled kwarg raises an error
        with pytest.raises(MISSING_ARG_ERRS):
            sel_data, tbool = select_time_range(data, time_range, time_axis=-1, timepts=timepts, foo=None)


@pytest.mark.parametrize('data_type',
                         [('spike_timestamp'),('spike_bool')])
def test_realign_data(spike_data, data_type):
    """ Unit tests for realign_data function """
    data, timepts = spike_data[data_type]
    data_orig = data.copy()
    n_trials, n_units = data.shape[:2]

    # Realign timestamps, then realign back to original timebase and test if same
    if data_type == 'spike_timestamp':
        realigned = realign_data(data, 0.5*np.ones((n_trials,)), trial_axis=0)
        realigned = realign_data(realigned, -0.5*np.ones((n_trials,)), trial_axis=0)
        assert object_array_equal(data,data_orig)     # Ensure input data not altered by func
        assert realigned.shape == data.shape
        for trial in range(n_trials):
            for unit in range(n_units):
                assert np.allclose(realigned[trial,unit], data[trial,unit])

        # Test for consistent output with transposed data dimensionality
        realigned = realign_data(data.T, 0.5*np.ones((n_trials,)), trial_axis=-1)
        realigned = realign_data(realigned, -0.5*np.ones((n_trials,)), trial_axis=-1)
        assert object_array_equal(data,data_orig)     # Ensure input data not altered by func
        assert realigned.shape == data.T.shape
        for trial in range(n_trials):
            for unit in range(n_units):
                assert np.allclose(realigned[unit,trial], data[trial,unit])

        # Ensure that passing a nonexistent/misspelled kwarg raises an error
        with pytest.raises(MISSING_ARG_ERRS):
            realigned = realign_data(data, 0.5*np.ones((n_trials,)), trial_axis=0, foo=None)

    # For boolean data, realign to 2 distinct times, then concatenate together and test if same
    else:
        realigned1 = realign_data(data, 0.5*np.ones((n_trials,)), time_range=(-0.5,-0.001),
                                  timepts=timepts, time_axis=-1, trial_axis=0)
        realigned2 = realign_data(data, 0.5*np.ones((n_trials,)), time_range=(0,0.5),
                                  timepts=timepts, time_axis=-1, trial_axis=0)
        realigned = np.concatenate((realigned1,realigned2), axis=-1)
        assert np.array_equal(data,data_orig)     # Ensure input data not altered by func
        assert realigned.shape == data.shape
        assert (realigned == data).all()

        # Test for consistent output with transposed data dimensionality
        realigned1 = realign_data(data.T, 0.5*np.ones((n_trials,)), time_range=(-0.5,-0.001),
                                  timepts=timepts, time_axis=0, trial_axis=-1)
        realigned2 = realign_data(data.T, 0.5*np.ones((n_trials,)), time_range=(0,0.5),
                                  timepts=timepts, time_axis=0, trial_axis=-1)
        realigned = np.concatenate((realigned1,realigned2), axis=0)
        assert np.array_equal(data,data_orig)     # Ensure input data not altered by func
        assert realigned.shape == data.T.shape
        assert (realigned == data.T).all()

        # Ensure that passing a nonexistent/misspelled kwarg raises an error
        with pytest.raises(MISSING_ARG_ERRS):
            realigned1 = realign_data(data.T, 0.5*np.ones((n_trials,)), time_range=(-0.5,-0.001),
                                            timepts=timepts, time_axis=0, trial_axis=-1, foo=None)


@pytest.mark.parametrize('data_type',
                         [('spike_timestamp'),('spike_bool')])
def test_pool_electrode_units(spike_data, data_type):
    """ Unit tests for pool_electrode_units function """

    def _count_all_spikes(data,elecs=None):
        if elecs is None: elecs = np.zeros(data.shape[1])
        elec_set = unsorted_unique(elecs)

        result = np.empty((len(elec_set,)),dtype=int)
        for i_elec,elec in enumerate(elec_set):
            if data_type == 'spike_bool':
                result[i_elec] = data[:,elecs==elec,:].sum()
            else:
                result[i_elec] = np.asarray([len(spikes) for spikes in data[:,elecs==elec].flatten()]).sum()

        return result

    data, timepts = spike_data[data_type]
    result = _count_all_spikes(data)
    # Concatenate data to simulate spiking data from 2 units on 2 electrodes
    data = np.concatenate((data,data), axis=1)
    data_orig = data.copy()
    electrodes = [1,1,2,2]
    n_trials, n_units = data.shape[:2]
    if data_type == 'spike_bool': n_timepts = data.shape[2]
    out_shape = (n_trials,2,n_timepts) if data_type == 'spike_bool' else (n_trials,2)

    data_checker = np.array_equal if data_type == 'bool' else object_array_equal

    # Check if overall spike count, size is consistent after pooling
    data_mua,elec_idxs = pool_electrode_units(data, electrodes, axis=1, return_idxs=True)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert data_mua.shape == out_shape
    assert (elec_idxs == np.array([0,2])).all()
    assert (_count_all_spikes(data,electrodes) == np.array([result,result])).all()

    # Check for consistent output with axis-transposed data
    data_mua,elec_idxs = pool_electrode_units(np.moveaxis(data,1,0), electrodes, axis=0,
                                              return_idxs=True)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func
    assert data_mua.shape == (out_shape[1],out_shape[0],*out_shape[2:])
    assert (elec_idxs == np.array([0,2])).all()
    assert (_count_all_spikes(data,electrodes) == np.array([result,result])).all()


# =============================================================================
# Unit tests for plotting functions
# =============================================================================
@pytest.mark.parametrize('data_type, graphics',
                         [('spike_timestamp',  'vector'),
                          ('spike_timestamp',  'bitmap'),
                          ('spike_bool',       'vector'),
                          ('spike_bool',       'bitmap')])
def test_plot_raster(spike_data, data_type, graphics):
    """ Unit tests for plot_raster function """
    # Extract given data type from data dict
    data, timepts = spike_data[data_type]
    data = data[:,0] if data_type == 'spike_timestamp' else data[:,0,:]
    data_orig = data.copy()

    data_checker = np.array_equal if data_type == 'bool' else object_array_equal

    extra_args = dict(graphics=graphics)
    if data_type == 'spike_bool':   extra_args.update(timepts=timepts)
    else:                           extra_args.update(lims=(0,1))

    # Basic test that call works
    ax = plot_raster(data, **extra_args)
    assert data_checker(data,data_orig)     # Ensure input data not altered by func

    # Test that a few parameter tweaks work
    ax = plot_raster(data, **extra_args, color=[0.25,0.25,0.25])
    ax = plot_raster(data, **extra_args, color='tab:gray')
    ax = plot_raster(data, **extra_args, height=0.75)
    ax = plot_raster(data, **extra_args,
                     events=[(100e-3,300e-3), 500e-3, (880e-3,900e-3,920e-3)])
    # Test with data from only one spike train
    ax = plot_raster(data=data[0] if data_type == 'spike_timestamp' else data[0,:], **extra_args)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        ax = plot_raster(data, **extra_args, foo=None)


def test_plot_mean_waveforms():
    """ Unit tests for plot_mean_waveforms function """
    n_units = 3
    n_spikes = 100
    n_timepts = 1000

    # Use dpss tapers as data to plot, since they are all very distinct looking
    means = compute_tapers(n_timepts, time_width=1.0, freq_width=4, n_tapers=n_units)
    waveforms = np.empty((n_units,), dtype=object)
    for unit in range(n_units):
        waveforms[unit] = means[:,[unit]] + 0.1*np.random.standard_normal((n_timepts,n_spikes))
    waveforms_orig = waveforms.copy()

    # Basic test that plotted data == input data
    lines, _, _ = plot_mean_waveforms(waveforms, plot_sd=True)
    assert object_array_equal(waveforms, waveforms_orig)
    for unit in range(n_units):
        assert np.allclose(lines[unit][0].get_ydata(), waveforms[unit].mean(axis=1))

    # Test w/o SD plot (mean only)
    lines, _, _ = plot_mean_waveforms(waveforms, plot_sd=False)
    assert object_array_equal(waveforms, waveforms_orig)
    for unit in range(n_units):
        assert np.allclose(lines[unit][0].get_ydata(), waveforms[unit].mean(axis=1))

    # Test with data from only one unit
    lines, _, _ = plot_mean_waveforms(waveforms[0], plot_sd=True)
    assert object_array_equal(waveforms, waveforms_orig)
    assert np.allclose(lines[0][0].get_ydata(), waveforms[0].mean(axis=1))

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        lines, _, _ = plot_mean_waveforms(waveforms, foo=None)


def test_plot_waveform_heatmap():
    """ Unit tests for plot_waveform_heatmap function """
    n_units = 1
    n_spikes = 100
    n_timepts = 1000

    # Use dpss tapers as data to plot, since they are all very distinct looking
    means = compute_tapers(n_timepts, time_width=1.0, freq_width=4, n_tapers=n_units)
    waveforms = np.empty((n_units,), dtype=object)
    for unit in range(n_units):
        waveforms[unit] = means[:,[unit]] + 0.1*np.random.standard_normal((n_timepts,n_spikes))
    waveforms_orig = waveforms.copy()

    # Basic test that plotted data == input data
    patch, _ = plot_waveform_heatmap(waveforms)
    assert object_array_equal(waveforms, waveforms_orig)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        patch, _ = plot_waveform_heatmap(waveforms, foo=None)


def test_imports():
    """ Test different import methods for spikes module """
    # Import entire package
    import spynal
    spynal.spikes.density
    # Import module
    import spynal.spikes as spk
    spk.density
    # Import specific function from module
    from spynal.spikes import density
    density



# -*- coding: utf-8 -*-
""" Oscillatory coherence analysis """
import numpy as np

from spynal.utils import axis_index_slices, setup_sliding_windows
from spynal.randstats.sampling import jackknifes
from spynal.randstats.utils import jackknife_to_pseudoval
from spynal.sync.helpers import _sync_raw_to_spectral, _sfc_raw_to_spectral


# =============================================================================
# Field-Field Synchrony functions
# =============================================================================
def coherence(data1, data2, axis=0, return_phase=False, transform=None, single_trial=None,
              spec_method='wavelet', data_type=None, smp_rate=None, time_axis=None,
              taper_axis=None, keepdims=True, **kwargs):
    """
    Compute coherence between pair of channels of raw or spectral (time-frequency) data
    (LFP or spikes)

    Coherence is a spectral analog of linear correlation that takes both phase and amplitude
    into account.

    Only parameters differing from :func:`.synchrony` are described here.

    Parameters
    ----------
    transform : 'Z' or None, default: None
        Transform to apply to all computed coherence values.
        Set=None to return raw, untransformed coherence.
        Set='Z' to Z-transform coherence using Jarvis & Mitra (2001) method.

    **kwargs :
        Any other keyword args passed as-is to spectrogram() function.
    """
    assert not((single_trial is not None) and return_phase), \
        ValueError("Cannot do both single_trial AND return_phase together")

    assert (single_trial is None) or (single_trial in ['pseudo','richter']), \
        ValueError("Parameter <single_trial> must be = None, 'pseudo', or 'richter'")

    if axis < 0: axis = data1.ndim + axis
    if (time_axis is not None) and (time_axis < 0): time_axis = data1.ndim + time_axis

    # Check if raw data input. If so, compute spectral transform first
    data1, data2, freqs, timepts, axis, time_axis, taper_axis = \
        _sync_raw_to_spectral(data1, data2, smp_rate, axis, time_axis, taper_axis,
                              spec_method, data_type, **kwargs)

    # For multitaper, compute means across trials, tapers; degrees of freedom = 2*n_trials*n_tapers
    if spec_method == 'multitaper':
        reduce_axes = (axis,taper_axis)
        df = 2*np.prod([data1.shape[ax] for ax in reduce_axes])
    # Otherwise just do means across trials; df = 2*n_trials
    else:
        reduce_axes = axis
        df = 2*data1.shape[axis]

    # Setup actual function call for any transform to compute on raw coherence values
    if (transform is None) or callable(transform):
        pass
    elif isinstance(transform,str):
        transform = transform.lower()    
        if transform in ['z','ztransform']:
            transform = lambda coh: ztransform_coherence(coh, df)
        else:
            raise ValueError("Unsupported value '%s' set for <transform>" % transform)
    else:
        raise TypeError("Unsupported type '%s' for <transform>. Use string or function or None" \
                        % type(transform))

    # Compute cross-spectrum and auto-spectrum of each channel
    auto_spec1 = data1*data1.conj()
    auto_spec2 = data2*data2.conj()
    cross_spec = data1*data2.conj()

    def _cross_auto_to_coh(cross_spec, auto_spec1, auto_spec2, axis,
                           return_phase, transform, keepdims):
        """ Compute coherence from cross-spectrum and pair of auto-spectra """
        # Average spectra across observatations (trials and/or tapers)
        # Note: .real deals with floating point error, converts complex dtypes to float
        auto_spec1 = np.mean(auto_spec1, axis=axis, keepdims=keepdims).real
        auto_spec2 = np.mean(auto_spec2, axis=axis, keepdims=keepdims).real
        cross_spec = np.mean(cross_spec, axis=axis, keepdims=keepdims)

        # Calculate complex coherency as cross-spectrum / product of square root of spectra
        coherency = cross_spec / np.sqrt(auto_spec1*auto_spec2)

        # Convert complex coherency -> coherence (magnitude of coherency)
        coherence = np.abs(coherency).real

        # Perform any requested tranform on coherence (eg z-scoring)
        if transform is not None: coherence = transform(coherence)

        # Optionally also extract coherence phase angle
        if return_phase:    return coherence, np.angle(coherency)
        else:               return coherence, None


    # Standard across-trial coherence estimator
    if single_trial is None:
        coh, dphi = _cross_auto_to_coh(cross_spec, auto_spec1, auto_spec2, reduce_axes,
                                        return_phase, transform, keepdims)

    # Single-trial coherence estimator using jackknife resampling method
    else:
        # Jackknife resampling of coherence statistic (this is the 'richter' estimator)
        # Note: Allow for reduction along taper axis within resampled stat function, but only
        #       resample across trial axis--want sync estimates for single-trials, not tapers
        jackfunc = lambda s12,s1,s2: _cross_auto_to_coh(s12, s1, s2, reduce_axes,
                                                        False, transform, True)[0]
        coh_shape = list(cross_spec.shape)
        if spec_method == 'multitaper': coh_shape[taper_axis] = 1
        n_jack = coh_shape[axis]
        ndim = len(coh_shape)

        # Create generator with n length-n vectors, each of which excludes 1 trial
        resamples = jackknifes(n_jack)

        # Do jackknife resampling -- estimate statistic w/ each observation left out
        coh = np.empty(coh_shape, dtype=float)
        for trial,sel in enumerate(resamples):
            # Index into <axis> of data and stat, with ':' for all other axes
            slices_in   = axis_index_slices(axis, sel, ndim)
            slices_out  = axis_index_slices(axis, [trial], ndim)
            coh[slices_out] = jackfunc(cross_spec[slices_in],
                                       auto_spec1[slices_in], auto_spec2[slices_in])

        # Convert to jackknife pseudovalues = n*stat_full - (n-1)*stat_jackknife
        # Note: n here is the number of jackknifes computed = n_trials
        if single_trial == 'pseudo':
            coh_full = jackfunc(cross_spec, auto_spec1, auto_spec2)
            coh = jackknife_to_pseudoval(coh_full, coh, data1.shape[axis])

        if not keepdims and (spec_method == 'multitaper'): coh = coh.squeeze(axis=taper_axis)

    if return_phase:    return coh, freqs, timepts, dphi
    else:               return coh, freqs, timepts


# =============================================================================
# Spike-Field Synchrony functions
# =============================================================================
def spike_field_coherence(spkdata, lfpdata, axis=0, time_axis=None, taper_axis=None,
                          timepts=None, transform=None,
                          data_type=None, spec_method='multitaper', smp_rate=None,
                          return_phase=False, keepdims=True, **kwargs):
    """
    Compute pairwise coherence between single-channel spiking data and LFP data

    Only parameters differing from :func:`.spike_field_coupling` are described here.

    Parameters
    ----------
    transform : 'Z' or None, default: None
        Transform to apply to all computed coherence values.
        Set=None to return raw, untransformed coherence.
        Set='Z' to Z-transform coherence using Jarvis & Mitra (2001) method.

    **kwargs :
        Any other keyword args passed as-is to spectrogram() function
    """
    if axis < 0: axis = lfpdata.ndim + axis
    if time_axis < 0: time_axis = lfpdata.ndim + time_axis

    # Check if raw data input. If so, compute spectral transform first
    spkdata, lfpdata, freqs, timepts, smp_rate, axis, time_axis, taper_axis = \
        _sfc_raw_to_spectral(spkdata, lfpdata, smp_rate, axis, time_axis, taper_axis, timepts,
                             'coherence', spec_method, data_type, **kwargs)

    extra_args = dict(axis=axis, data_type='spectral', spec_method=spec_method,
                      transform=transform, return_phase=return_phase)
    if spec_method == 'multitaper': extra_args.update(taper_axis=taper_axis)

    if return_phase:
        coh,_,_,phi = coherence(spkdata, lfpdata, keepdims=keepdims, **extra_args)
        return coh,freqs,timepts,None,phi
    else:
        coh,_,_ = coherence(spkdata, lfpdata, keepdims=keepdims, **extra_args)
        return coh,freqs,timepts,None


# =============================================================================
# Helper functions
# =============================================================================
def ztransform_coherence(coh, df, beta=23/20):
    """
    z-transform coherence values to render them approximately normally distributed

    Parameters
    ----------
    coh : ndarray, shape=Any
        Raw coherence values

    df : int
        Degrees of freedom of coherence estimates.
        For multitaper spectral estimates, this is usually df = 2*n_trials*n_tapers.
        For other estimates, this is usually df = 2*n_trials.

    beta : scalar, default: 23/20
        Mysterious number from Jarvis & Mitra to make output z-scores approximately normal

    Returns
    -------
    z : ndarray, shape=Any
        z-transformed coherence. Same shape as `coh`.

    References
    ----------
    - Jarvis & Mitra (2001) Neural Computation https://doi.org/10.1162/089976601300014312
    - Hipp, Engel, Siegel (2011) Neuron https://doi.org/10.1016/j.neuron.2010.12.027
    """
    return beta*(np.sqrt(-(df-2)*np.log(1-coh**2)) - beta)




# -*- coding: utf-8 -*-
""" Preprocessing functions for LFP/EEG/continuous data and spectral analysis """
import numpy as np

from sklearn.linear_model import LinearRegression

from spynal.helpers import _check_window_lengths
from spynal.utils import index_axis, standardize_array, undo_standardize_array


def cut_trials(data, trial_lims, smp_rate, axis=0):
    """
    Cut continuous (eg LFP) data into trials

    Parameters
    ----------
    data : ndarray, shape=(...,n_timepts,...)
        Continuous data unsegmented into trials.
        Arbitrary dimensionality, could include multiple channels, etc.

    trial_lims : array-like, shape=(n_trials,2)
        List of [start,end] of each trial (in s) to use to cut data.

    smp_rate : scalar
        Sampling rate of data (Hz).

    axis : int, default: 0 (1st axis)
        Axis of data array corresponding to time samples

    Returns
    -------
    cut_data : ndarray, shape=(...,n_trial_timepts,...,n_trials)
        Continuous data segmented into trials.
        Trial axis is appended to end of all axes in input data.
    """
    trial_lims = np.asarray(trial_lims)
    assert (trial_lims.ndim == 2) and (trial_lims.shape[1] == 2), \
        "trial_lims argument should be a (n_trials,2) array of trial [start,end] times"
    n_trials = trial_lims.shape[0]

    # Convert trial_lims in s -> indices into continuous data samples
    trial_idxs = np.round(smp_rate*trial_lims).astype(int)
    assert trial_idxs.min() >= 0, \
        ValueError("trial_lims are attempting to index before start of data")
    assert trial_idxs.max() < data.shape[axis], \
        ValueError("trial_lims are attempting to index beyond end of data")
    # Ensure all windows have same length
    trial_idxs = _check_window_lengths(trial_idxs,tol=1)

    # Samples per trial = end - start + 1
    n_smp_per_trial = trial_idxs[0,1] - trial_idxs[0,0] + 1

    # Create array to hold trial-cut data. Same shape as data, with time sample axis
    # reduced to n_samples_per_trial and trial axis appended.
    cut_shape = [*data.shape,n_trials]
    cut_shape[axis] = n_smp_per_trial
    cut_data = np.empty(tuple(cut_shape),dtype=data.dtype)

    # Extract segment of continuous data for each trial
    for trial,lim in enumerate(trial_idxs):
        cut_data[...,trial] = index_axis(data, axis, slice(lim[0],lim[1]+1))

    return cut_data


def realign_data(data, align_times, time_range, timepts, time_axis=0, trial_axis=-1):
    """
    Realigns trial-cut continuous (eg LFP) data to new set of within-trial times
    (eg new trial event) so that t=0 on each trial at given event.
    For example, data aligned to a start-of-trial event might
    need to be relaligned to the behavioral response.

    Parameters
    ----------
    data : ndarray, shape=(...,n_timepts,...,n_trials,...)
        Continuous data segmented into trials.
        Arbitrary dimensionality, could include multiple channels, etc.

    align_times : array-like, shape=(n_trials,)
        New set of times (in old reference frame) to realign data to (in s)

    time_range : array-like, shape=(2,)
        Time range to extract from each trial around new align time
        ([start,end] in s relative to align_times).
        eg, time_range=(-1,1) -> extract 1 s on either side of align event.

    timepts : array-like, shape=(n_timepts)
        Time sampling vector for data (in s)

    time_axis : int, default: 0 (1st axis of array)
        Axis of data corresponding to time samples

    trial_axis : int, default: -1 (last axis of array)
        Axis of data corresponding to distinct trials

    Returns
    -------
    realigned : ndarray, shape=(...,n_timepts_out,...,n_trials,...)
        Data realigned to given within-trial times.
        Time axis is reduced to length implied by `time_range`, but otherwise
        array has same shape as input data.
    """
    assert time_range is not None, \
        "Desired time range to extract from each trial must be given in  `time_range`"
    assert timepts is not None, "Data time sampling vector must be given in `timepts`"

    timepts     = np.asarray(timepts)
    align_times = np.asarray(align_times)
    time_range  = np.asarray(time_range)

    if time_axis < 0:   time_axis = data.ndim + time_axis
    if trial_axis < 0:  trial_axis = data.ndim + trial_axis

    # Move array axes so time axis is 1st and trials last (n_timepts,...,n_trials)
    if (time_axis == data.ndim-1) and (trial_axis == 0):
        data = np.swapaxes(data,time_axis,trial_axis)
    else:
        if time_axis != 0:              data = np.moveaxis(data,time_axis,0)
        if trial_axis != data.ndim-1:   data = np.moveaxis(data,trial_axis,-1)

    # Convert align times and time epochs to nearest integer sample indexes
    dt = np.mean(np.diff(timepts))
    align_smps = np.round((align_times - timepts[0])/dt).astype(int)
    range_smps = np.round(time_range/dt).astype(int)
    # Compute [start,end] sample indexes for each trial epoch = align time +/- time range
    trial_range_smps = align_smps[:,np.newaxis] + range_smps[np.newaxis,:]

    assert (trial_range_smps[:,0] >= 0).all(), \
        "Some requested time epochs extend before start of data"
    assert (trial_range_smps[:,1] < len(timepts)).all(), \
        "Some requested time epochs extend beyond end of data"

    n_timepts_out   = range_smps[1] - range_smps[0] + 1
    return_shape    = (n_timepts_out, *(data.shape[1:]))
    realigned       = np.empty(return_shape)

    # Extract timepoints corresponding to realigned time epoch from each trial in data
    for trial,t in enumerate(trial_range_smps):
        # Note: '+1' below makes the selection inclusive of the right endpoint in each trial
        realigned[...,trial] = data[t[0]:t[1]+1,...,trial]

    # Move array axes back to original locations
    if (time_axis == data.ndim-1) and (trial_axis == 0):
        realigned = np.swapaxes(realigned,trial_axis,time_axis)
    else:
        if time_axis != 0:              realigned = np.moveaxis(realigned,0,time_axis)
        if trial_axis != data.ndim-1:   realigned = np.moveaxis(realigned,-1,trial_axis)

    return realigned


def realign_data_on_event(data, event_data, event, timepts, align_times, time_range,
                          time_axis=0, trial_axis=-1):
    """
    Convenience wrapper around `realign_data` for relaligning to a given
    named event within a per-trial dataframe or dict variable.

    Only parameters differing from :func:`realign_data` are described here.

    Parameters
    ----------
    event_data : dict, {str : ndarray, shape=(n_trials,)} or DataFrame, shape=(n_trials,n_events)
        Per-trial event timing data to use to realign spike timestamps.

    event : str
        Dict key or DataFrame column name whose associated values are to be used to realign data
    """
    # Extract vector of times to realign on
    align_times = event_data[event]

    # Compute the realignment and return
    return realign_data(data, timepts, align_times, time_range,
                        time_axis=time_axis, trial_axis=trial_axis)


def remove_dc(data, axis=None):
    """
    Remove constant DC component of signals, estimated as across-time mean
    for each time series (ie trial,channel,etc.)

    Parameters
    ----------
    data : ndarray, shape=(...,n_timepoints,...)
        Raw data to remove DC component of.
        Can be any arbitary shape, with time sampling along `axis`

    axis : int, Default: None (remove DC component computed across *full* data array)
        Data axis corresponding to time

    Returns
    -------
    data : ndarray, shape=(...,n_timepoints,...)
        Data with DC component removed (same shape as input)
    """
    return data - data.mean(axis=axis, keepdims=True)


def remove_evoked(data, axis=0, method='mean', design=None, return_evoked=False):
    """
    Remove estimate of evoked potentials phase-locked to trial events,
    returning data with (in theory) only non-phase-locked induced components

    Parameters
    ----------
    data : ndarray, shape=(...,n_obs,...)
        Raw data to remove evoked components from.
        Can be any arbitary shape, with observations (trials) along `axis`.

    axis : int, default: 0 (1st axis)
        Data axis corresponding to distinct observations/trials

    method : {'mean','groupmean','regress'}, default: 'mean'
        Method to use for estimating evoked potentials:

        - 'mean'      : Grand mean signal across all observations (trials)
        - 'groupmean' : Mean signal across observations with each group in `design`
        - 'regress'   : OLS regresion fit of design matrix `design` to data

    design : array-like, shape=(n_obs,...), optional
        Design matrix to fit to data (`method` == 'regress')
        or group/condition labels for each observation (`method` == 'groupmean').
        Not used for `method` == 'mean'.

    return_evoked : bool, default: False
        If True, also returns second output = estimated evoked potential on each trial.

    Returns
    -------
    data_induced : ndarray, shape=(...,n_obs,...)
        Data with estimated evoked component removed, leaving only non-phase-locked
        "induced" component. Same shape as input `data`.

    evoked : ndarray, shape=(...,n_obs,...), optional
        Evoked potential for each trial, estimated with given method. Same shape as `data`.
        Only returned if `return_evoked` is True.
    """
    assert axis is not None, "<axis> should correspond to data trials/observations dimension"
    
    method = method.lower()
    design = np.asarray(design)

    data = data.copy()  # Copy input data to avoid overwriting it in caller
    
    # Subtract off grand mean potential across all trials
    if method in ['mean','grandmean']:
        evoked_grandmean = np.mean(data, axis=axis, keepdims=True)
        data -= evoked_grandmean
        
        # Expand grand-mean evoked potential to n_trials if returning evoked
        if return_evoked:
            reps = np.ones((data.ndim,),dtype=int)
            reps[axis] = data.shape[axis]
            evoked = np.tile(evoked_grandmean, reps)

    # Subtract off mean potential across all trials within each group/condition
    # todo  can we do this with an xarray or pandas groupby() instead??
    elif method == 'groupmean':
        assert (design.ndim == 1) or ((design.ndim == 2) and (design.shape[1] == 1)), \
            "Design matrix <design> must be vector-like (1d or 2d w/ shape[1]=1)"

        groups = np.unique(design)

        data, data_shape = standardize_array(data, axis=axis, target_axis=0)
        if return_evoked: evoked = np.empty_like(data)
        
        for group in groups:
            idxs = design == group
            evoked_group = np.mean(data[idxs,...], axis=0, keepdims=True)
            data[idxs,...] -= evoked_group
            if return_evoked: evoked[idxs,...] = evoked_group

        data = undo_standardize_array(data, data_shape, axis=axis, target_axis=0)
        if return_evoked:
            evoked = undo_standardize_array(evoked, data_shape, axis=axis, target_axis=0)

    # Regress data on given design matrix and return residuals
    elif method == 'regress':
        assert design.ndim in [1,2], \
            "Design matrix <design> must be matrix-like (2d) or vector-like (1d)"

        # Only fit explicit intercept if design doesn't already have a constant (intercept) column in it
        has_constant_col = np.any(np.all(design==1, axis=0))
        model = LinearRegression(fit_intercept=True if not has_constant_col else False)

        data, data_shape = standardize_array(data, axis=axis, target_axis=0)

        evoked = model.fit(design,data).predict(design)
        data -= evoked

        data = undo_standardize_array(data, data_shape, axis=axis, target_axis=0)
        if return_evoked:
            evoked = undo_standardize_array(evoked, data_shape, axis=axis, target_axis=0)

    if return_evoked:   return data, evoked
    else:               return data

